function [w, b, obj_val, pd_gap] = consensus_pdhg(X, y, lambda, max_iter, tol)
%CONSENSUS_PDHG Summary of this function goes here
%   Detailed explanation goes here

    [N, d] = size(X);

    
    pd_gap = zeros(1, max_iter);
    
    % setup linear operator
    K = repmat(speye(d+1), N, 1);
    
    norm_K = normest(K);
    
    tau = 1/norm_K;
    sigma = 1/norm_K;
    
    % init dual variable
    p = zeros(N*(d+1), 1);
    
    % init primal variable
    u = zeros(d+1, 1);
    u_bar = u;
    
    % assemble matrix A \in R^{d+1 x N} so that <(w,b), (x_i, 1)*y_i> = (<w, x_i> + b)*y_i
    A = [X'; ones(1, N)];
    A = A .* repmat(y', d+1, 1);
    
    sq_norm_a = diag(A'*A);
    
    primal_handle = @(u) sum(max(0, 1 - A'*u)) + lambda*sum(abs(u(1:end-1)));
    obj_val = zeros(1, max_iter);
    for it=1:max_iter
        p = reshape(p + sigma*(K*u_bar), d+1, N);
        
        % prox_F^*
        s = (diag(A'*p) - sigma) ./ sq_norm_a;
        s(s < -1) = -1;
        s(s > 0) = 0;
        p = A.*repmat(s', d+1, 1);
        p = p(:);

        r = -K'*p;
        relaxed_dual_obj = -sum(s) - r(end)^2 - sum(max(0, r(1:end-1)-lambda).^2);
        
        u_bar = -u;
        u = u - tau*(K'*p);
        % prox_G 
        % apply shrinkage operator to w only! Don't regularize the bias.
        u(1:end-1) = max(abs(u(1:end-1))-tau*lambda, 0).*sign(u(1:end-1));      
        
        u_bar = u_bar + 2*u;
        obj_val(it) = primal_handle(u);
        pd_gap(it) = primal_handle(u) - relaxed_dual_obj;
        
        if(abs(pd_gap(it)) < tol)
            break;
        end
        
        % print objective value and relaxed primal-dual gap
        fprintf('It=%d    Objective=%f    Primal-Dual-Gap=%f\n', it, obj_val(it), pd_gap(it));
    end
    
    w = u(1:d);
    b = u(end);
end

